/*
	Prendre ta douche !

    Les douches des garçons du gymnase sont en travaux. Pendant ce temps, les garçons et les
filles utiliseront les douches des filles. Afin de ne choquer personne,  un panneau est accroché
à l'entrée des douches : libre, occuppé par des filles ou occuppé par des garçons. Seules des
filles peuvent entrer dans les douches lorsque des filles s'y trouvent déja et réciproquement
pour les garçons.

    Créez un programme simulant un groupe de G garçons et un groupe de F filles attendant de
prendre leur douche. Chaque personne sera simulée par un thread, et D cabines de douches
seront disponibles.

    Existe-t-il une solution plus efficace que les autres ? En effet, il est possible de donner la
priorité à un groupe (les filles d'abord, par exemple), cette "stratégie" est-elle plus efficace
qu'une solution équitable ? Mesurez le temps nécessaire pour que tout le monde se douche
en fonction de la stratégie utilisée.

 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <pthread.h>

// nombre de personnes
#define N 50

// nombre de cabines de douches
#define D 4

// verrou pour l'exclusion
typedef struct
{
	int filles_n;
	pthread_mutex_t filles_mutex;

	int garcons_n;
	pthread_mutex_t garcons_mutex;

	int cabines_libres_n;
	pthread_cond_t cabine_libre;
} MUTEX_FG;

// description d'une personne
typedef struct
{
	char prenom[16];
	char sexe;

	MUTEX_FG *mutex;	// toutes les personnes partagent le meme verrou

	int heure_douche;	// pour les statistiques
	int duree_douche;

	int duree_attente;
	int heure_fin_douche;
} PERSONNE;



// intialisation du verrou
MUTEX_FG *fg_init(MUTEX_FG *m)
{
	pthread_mutex_init(&m->filles_mutex, NULL);
	pthread_mutex_init(&m->garcons_mutex, NULL);
	pthread_cond_init(&m->cabine_libre, NULL);

	m->filles_n= 0;
	m->garcons_n= 0;

	m->cabines_libres_n= D;

	return m;
}

/* section critique d'un garcon
 */

// entree
void fg_garcon_debut_douche(MUTEX_FG *m)
{
	pthread_mutex_lock(&m->garcons_mutex);	// bloque les autres garcons	
	while(m->cabines_libres_n<=0)
		// attend qu'une cabine soit disponible
		pthread_cond_wait(&m->cabine_libre, &m->garcons_mutex);

	m->garcons_n++;	// compte le nombre de garcon en train de se doucher
	if(m->garcons_n==1)
	{
		// bloque les filles
		pthread_mutex_lock(&m->filles_mutex);
	}

	// une cabine est occuppee par le garcon
	m->cabines_libres_n--;

	pthread_mutex_unlock(&m->garcons_mutex);
}

// sortie
void fg_garcon_fin_douche(MUTEX_FG *m)
{
	pthread_mutex_lock(&m->garcons_mutex);

	m->garcons_n--;	// compte le nombre de garcons en train de se doucher
	if(m->garcons_n==0)
	{
		// de-bloque les filles, si c'est le dernier garcon a se doucher
		pthread_mutex_unlock(&m->filles_mutex);
	}

	// signale qu'une cabine est disponible
	m->cabines_libres_n++;
	pthread_cond_broadcast(&m->cabine_libre);

	pthread_mutex_unlock(&m->garcons_mutex);
}



/* section critique d'une fille / symetrique par rapport aux garcons
   il serait souhaitable de n'utiliser que 2 fonctions d'entree / sortie de section critique
   (avec un parametre supplementaire) au lieu des 4 utilisees dans cette solution.
 */

// entree
void fg_fille_debut_douche(MUTEX_FG *m)
{
	pthread_mutex_lock(&m->filles_mutex);

	while(m->cabines_libres_n<=0)
		// attend qu'une cabine soit disponible
		pthread_cond_wait(&m->cabine_libre, &m->filles_mutex);

	m->filles_n++;
	if(m->filles_n==1)
	{
		// bloque les garcons
		pthread_mutex_lock(&m->garcons_mutex);
	}

	// une cabine est occuppee par la fille
	m->cabines_libres_n--;

	pthread_mutex_unlock(&m->filles_mutex);
}

// sortie
void fg_fille_fin_douche(MUTEX_FG *m)
{
	pthread_mutex_lock(&m->filles_mutex);

	m->filles_n--;
	if(m->filles_n==0)
	{
		// de-bloque les garcons
		pthread_mutex_unlock(&m->garcons_mutex);
	}

	// signale qu'une cabine est disponible
	m->cabines_libres_n++;
	pthread_cond_broadcast(&m->cabine_libre);

	pthread_mutex_unlock(&m->filles_mutex);
}



/* threads de simulation
 */

// simulation d'un garcon
void *garcon(void *voidparam)
{
	PERSONNE *p= (PERSONNE *) voidparam;

	// fait la queue
	printf("%s attend...\n", p->prenom);
	fg_garcon_debut_douche(p->mutex);

	// prend sa douche
	p->duree_attente= time(NULL) - p->heure_douche;
	printf("  %s prend sa douche (%d secondes, attente %ds)\n", p->prenom, p->duree_douche, p->duree_attente);
	sleep(p->duree_douche);

	// repart
	printf("    %s est propre\n", p->prenom);
	fg_garcon_fin_douche(p->mutex);

	printf("%s repart\n", p->prenom);
	p->heure_fin_douche= time(NULL);

	pthread_exit(NULL);
	return NULL;
}


// simulation d'une fille
void *fille(void *voidparam)
{
	PERSONNE *p= (PERSONNE *) voidparam;

	// fait la queue
	printf("%s attend...\n", p->prenom);
	fg_fille_debut_douche(p->mutex);

	// prend sa douche
	p->duree_attente= time(NULL) - p->heure_douche;
	printf("  %s prend sa douche (%d secondes, attente %ds)\n", p->prenom, p->duree_douche, p->duree_attente);
	sleep(p->duree_douche);

	// repart
	printf("    %s est propre\n", p->prenom);
	fg_fille_fin_douche(p->mutex);

	printf("%s repart\n", p->prenom);
	p->heure_fin_douche= time(NULL);

	pthread_exit(NULL);
	return NULL;
}


/*
 */

int main(void)
{
	PERSONNE p[N];
	pthread_t tid[N];
	MUTEX_FG verrou;
	float tmp;
	int t;
	int heure_debut, heure_fin;
	int code;
	int i;

	srand(time(NULL));

	// creer le verrou
	fg_init(&verrou);


	printf("douche : %d personnes, %d cabines de douches\n", N, D);

	// creation des filles et des garcons
	heure_debut= time(NULL);
	for(i= 0; i < N; i++)
	{
		// attendre, tout le monde ne se presente pas en meme temps
		sleep(rand() % 2);

		if(rand() < RAND_MAX/2)
		{
			p[i].mutex= &verrou;	// toutes les personnes partagent le meme verrou
			p[i].sexe= 'G';
			sprintf(p[i].prenom, "ken%d", i);
			p[i].duree_douche= 3 + rand() % 5; // entre 3 et 7 secondes pour se doucher
			p[i].heure_douche= time(NULL);	// heure d'arrive pour prendre la douche

			code= pthread_create(&tid[i], NULL, garcon, &p[i]);
		}
		else
		{
			p[i].mutex= &verrou;	// toutes les personnes partagent le meme verrou
			p[i].sexe= 'F';
			sprintf(p[i].prenom, "barbie%d", i);
			p[i].duree_douche= 3 + rand() % 5; // entre 3 et 7 secondes pour se doucher
			p[i].heure_douche= time(NULL);	// heure d'arrive pour prendre la douche

			code= pthread_create(&tid[i], NULL, fille, &p[i]);
		}
	}

	// attendre
	for(i= 0; i < N; i++)
		code= pthread_join(tid[i], NULL);

	heure_fin= time(NULL);


	/* affiche quelques statistiques
	 */
	printf("\n%d cabines disponibles, %d personnes\n", D, N);
	printf("%d secondes\n", heure_fin - heure_debut);

	tmp= 0.;
	for(i= 0; i < N; i++)
		tmp+= p[i].duree_douche;

	printf("duree moyenne d'une douche %f\n", tmp / N);

	tmp= 0.;
	for(i= 0; i < N; i++)
		tmp+= p[i].duree_attente;

	printf("duree d'attente moyenne %f\n", tmp / N);

	//
	t= p[0].heure_douche;
	for(i= 1; i < N; i++)
		if(p[i].heure_douche < t)
			t= p[i].heure_douche;

	printf("premiere arrivee h+%d\n", t - heure_debut);

	tmp= p[0].heure_douche;
	for(i= 1; i < N; i++)
		if(p[i].heure_douche > t)
			t= p[i].heure_douche;

	printf("derniere arrivee h+%d\n", t - heure_debut);

	//
	t= p[0].duree_attente;
	for(i= 1; i < N; i++)
		if(p[i].duree_attente < t)
			t= p[i].duree_attente;

	printf("plus courte attente %d\n", t);

	t= p[0].duree_attente;
	for(i= 1; i < N; i++)
		if(p[i].duree_attente > t)
			t= p[i].duree_attente;

	printf("plus longue attente %d\n", t);


	return 0;
}

